#include <ZUNO_LiquidCrystal.h>
#include <SPI.h>
#include <ZUNO_MFRC522.h>

#define LCD_RS 9
#define LCD_EN 10
#define LCD_D4 11
#define LCD_D5 12
#define LCD_D6 21
#define LCD_D7 22
#define RFID_SDA 6
#define RFID_SCK 0
#define RFID_MISO 1
#define RFID_MOSI 2
#define RFID_RST 7

#define BLUEPIN 18
#define YELLOWPIN  19
#define GREENPIN  20
#define REDPIN  24

#define NOPIN  255 // This should always equal an invalid pin number

#define NR_KNOWN_UIDS  5

#define MENU1             "House  Garage SR"
#define MENU2_INITIAL     " |     |      | "
#define MENU2_OLARMAT     " |     |      |*"
#define MENU2_SHELL       " |     |*     | "
#define MENU2_LARMAT      " |*    |      | "
#define NOTICE1_LARMAT    "Alarm activated "
#define NOTICE2_LARMAT    "Have a nice day "
#define NOTICE1_SHELL     "Shell activated "
#define NOTICE2_SHELL     "Goodnight       "
#define NOTICE1_OLARMAT   "Alarm disabled  "
#define NOTICE2_OLARMAT   "Welcome home    "
#define NOTICE1_AUTHORIZE "Authorizing...  "
#define NOTICE2_AUTHORIZE "Waiting for key "

unsigned long lastDebounceTime = 0;  // the last time the output pin was toggled
unsigned long debounceDelay = 50;    // the debounce time; increase if the output flickers

unsigned long authorizeStart = 0;
unsigned long authorizeDelay = 5*1000; // 5 seconds

struct Uid knownUIDs[NR_KNOWN_UIDS] = {
  {0x0A,  {0x96, 0x81, 0x53, 0x14, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},  0x00},
  {0x0A, {0x34, 0x9A, 0x96, 0xBB, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},  0x00},
  {0x0A, {0x77, 0xBF, 0x96, 0xBB, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},  0x00},
  {0x0A, {0xD9, 0x38, 0x99, 0x5B, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},  0x00},
  {0x0A, {0xB6, 0xC4, 0x95, 0xBB, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},  0x00}
};

LiquidCrystal lcd(LCD_RS, LCD_EN, LCD_D4, LCD_D5, LCD_D6, LCD_D7);
MFRC522 mfrc522(RFID_SDA, RFID_RST);   // Create MFRC522 instance.

enum state {
  initial,
  larmat,
  shell,
  olarmat,
};
  
enum state current_state = initial;
enum state target_state = initial;

void setup() 
{
  Serial.begin(9600);
  Serial.println("setup");
  pinMode(GREENPIN, INPUT);
  pinMode(YELLOWPIN, INPUT);
  pinMode(REDPIN, INPUT);

  pinMode(RFID_SDA, OUTPUT);

  mfrc522.PICC_DebugSerial(true);
}

void lcd_print(char* firstLine, char* secondLine)
{
  lcd.setCursor(0,0);
  lcd.print(firstLine);
  lcd.setCursor(0,1);
  lcd.print(secondLine);
} 

void dump_byte_array(byte *buffer, byte bufferSize) {
  for (byte i = 0; i < bufferSize; i++) {
      if (buffer[i] < 0x10)
        Serial.print(" 0");
      else
        Serial.print(" ");
      Serial.print(buffer[i], HEX);
  }
}

void loop() 
{
  // if no state change required...
  if (current_state == target_state) {
//      Serial.println ("Check button pressed");
    int pin = check_button_pressed();
    switch (pin) {
      case REDPIN:
        Serial.println ("Red button pressed");
        set_target_larmat();
      break;

      case YELLOWPIN:
        Serial.println ("Yellow button pressed");
        set_target_shell();
      break;

      case GREENPIN:
        Serial.println ("Green button pressed");
        set_target_olarmat();
      break;

      case BLUEPIN:
        Serial.println ("Blue button pressed");
        lcd.begin(16, 2);
        lcd.clear();
        setInitialMenu();
        SPI.begin();          // Init SPI bus
        mfrc522.PCD_Init(RFID_SDA, RFID_RST);   // Init MFRC522 card
/*        if (mfrc522.PCD_PerformSelfTest())
          Serial.println ("Self test passed");
        else
          Serial.println ("Self test failed");
*/
      break;

      default:
      break;
    }
  }
  else {
    // if we haven't told the user yet that they need to authorize, do so now
    if (authorizeStart == 0) {
      authorizeStart = millis();
      lcd_print (NOTICE1_AUTHORIZE,NOTICE2_AUTHORIZE);
      Serial.println ("Authorize start...");
    }

    if (is_authorized()) {
      transition_to_target();
      authorizeStart = 0;
    }
    else {
//      Serial.println ("Checking for timeout...");
      // if we've timed out waiting for authorization, go back to the current state (set target_state = current_state)
      if (authorizeStart + authorizeDelay < millis()) {
        revert_to_current();
        authorizeStart = 0;
      }
    }
  }
}

void revert_to_current()
{
  target_state = current_state;
  switch (current_state) {
    case initial:
      setInitialMenu();
    break;

    case larmat:
      setLarmatMenu();
    break;

    case shell:
      setShellMenu();
    break;

    case olarmat:
      setOlarmatMenu();
    break;
  }
}

void transition_to_target()
{
  switch (target_state) {
    case initial:
      Serial.println("Invalid target state");
      lcd_print ("Invalid target state","");
    break;

    case larmat:
      setLarmat();
    break;

    case shell:
      setShell();
    break;

    case olarmat:
      setOlarmat();
    break;
  }
}

bool try_key(Uid* uid)
{
  for (int i = 0; i < NR_KNOWN_UIDS; i++) {
    if(memcmp(uid->uidByte, knownUIDs[i].uidByte, mfrc522.m_uid.size) == 0)
    {
      Serial.print("GOOD KEY: ");
      dump_byte_array(uid->uidByte, uid->size);
      Serial.println();
      return true;
    }
  }
    
  Serial.print("BAD KEY: ");
  dump_byte_array(uid->uidByte, uid->size);
  Serial.println();
  return false;
}

bool is_authorized ()
{
//  Serial.println("CheckNewCard");
  if (!mfrc522.PICC_IsNewCardPresent())
     return false;

  // Select one of the cards
  Serial.println("ReadCardSerial");
  if (!mfrc522.PICC_ReadCardSerial())
    return false;

  Serial.print("Card ID:");
  dump_byte_array(mfrc522.m_uid.uidByte, mfrc522.m_uid.size);
  Serial.println();
  Serial.print("PICC type: ");
  byte piccType = mfrc522.PICC_GetType(mfrc522.m_uid.sak);
  Serial.println();

  // Try the key
  if (try_key(&mfrc522.m_uid) == true) {
    return true;
  }
  return false;
}

int check_button_pressed()
{
  int currentStateGreen = digitalRead(GREENPIN);
  int currentStateYellow = digitalRead(YELLOWPIN);
  int currentStateRed = digitalRead(REDPIN);
  int currentStateBlue = digitalRead(BLUEPIN);

  if (currentStateGreen == HIGH)
    return GREENPIN;
  
  if (currentStateYellow == HIGH)
    return YELLOWPIN;

  if (currentStateRed == HIGH)
    return REDPIN;

  if (currentStateBlue == HIGH)
    return BLUEPIN;

  return NOPIN;
}

void set_target_larmat()
{
  target_state = larmat;
}

void set_target_shell()
{
  target_state = shell;
}

void set_target_olarmat()
{
  target_state = olarmat;
}

void setLarmat() {
  Serial.println("Set armed");
  current_state = larmat;
  // updateZWaveStatus();
  setLarmatMenu();
}

void setLarmatMenu() {
  Serial.println("Set armed menu");
  lcd_print(NOTICE1_LARMAT,NOTICE2_LARMAT);
  delay(3000);
  lcd_print (MENU1,MENU2_LARMAT);
}

void setShell() {
  Serial.println("Set shell");
  current_state = shell;
  // updateZWaveStatus();
  setShellMenu();
}

void setShellMenu() {
  Serial.println("Set shell menu");
  lcd_print(NOTICE1_SHELL,NOTICE2_SHELL);
  delay(3000);
  lcd_print (MENU1,MENU2_SHELL);
}

void setOlarmat() {
  Serial.println("Set armed");
  current_state = olarmat;
  // updateZWaveStatus();
  setOlarmatMenu();
}

void setOlarmatMenu() {
  Serial.println("Set armed menu");
  lcd_print(NOTICE1_OLARMAT,NOTICE2_OLARMAT);
  delay(3000);
  lcd_print (MENU1,MENU2_OLARMAT);
}

void setInitialMenu() {
  Serial.println("Set initial menu");
  lcd_print(MENU1, MENU2_INITIAL);
}

